using System;
using System.Collections;
using System.Runtime.InteropServices;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Reflection;
using System.Runtime.CompilerServices;

// To get the imports, add dependencys to the files 
// OpcRcwComn.dll and OpcRcw.Da.dll typically located in the
// <Program Files>\<Common Files>\OPC Foundation\Bin folder
using OpcRcw.Comn;
using OpcRcw.Da;

namespace IBHOpc_CSharp
{
    public partial class Form1 : Form, IOPCDataCallback
    {
        public Form1()
        {
            InitializeComponent();
        }

        // Define the names
        internal const string IBH_OPCSERVER_NAME = "IBHsoftec.IBHOPC.DA.1";     // IBH OPC Server
        internal const string IBH_ITEM1_NAME = "Steuerung1.MW0";                // Item #1
        internal const string IBH_ITEM2_NAME = "Steuerung1.MW2";                // Item #2
        internal const string IBH_GROUP_NAME = "MyIBHOPCGroup1";                // OPC Group
        internal const int iLocale_ID = 0x407;                                  // English.

        // Globals
        IOPCServer pIBHOPCServer;
        IOPCAsyncIO2 pIOPCAsyncIO2 = null;                                      // Asynchronous IO.
        IOPCGroupStateMgt pIOPCGroupStateMgt = null;
        IConnectionPointContainer pIConnectionPointContainer = null;
        IConnectionPoint pIConnectionPoint = null;

        Object pIBHGroup = null;
        int pServerGroupHandle = 0;                                             // Group handle
        int iTransaction_ID = 0;
        int[] ServerItem_HandleArray;
        Int32 dwCookie = 0;

        private void StartIBHOpc_Click(object sender, EventArgs e)
        {
            // Local variables
            Type IBHServer;
            OPCITEMDEF[] ItemDefinitionArray;

            // Group properties
            Int32 dwRequestedUpdateRate = 1000;
            Int32 hClientGroup = 1;
            Int32 pRevisedUpdateRate;
            float fDeadband = 0;
            int TimeBias = 0;

            GCHandle hTimeBias, hDeadband;
            hTimeBias = GCHandle.Alloc(TimeBias, GCHandleType.Pinned);
            hDeadband = GCHandle.Alloc(fDeadband, GCHandleType.Pinned);

            // Create the IBH OPC Server
            Guid iidRequiredInterface = typeof(IOPCItemMgt).GUID;
            IBHServer = Type.GetTypeFromProgID(IBH_OPCSERVER_NAME);
            try
            {
                // Create the IBH OPC Server object
                pIBHOPCServer = (IOPCServer)Activator.CreateInstance(IBHServer);
                try
                {
                    // Add the group "MyIBHOPCGroup"
                    // [in] not active, so no OnDataChange callback
                    // [in] Request this Update Rate from Server
                    // [in] Client Handle, not necessary in this sample
                    // [in] No time interval to system UTC time
                    // [in] No Deadband, so all data changes are reported
                    // [in] Server uses english language to for text values
                    // [out] Server handle to identify this group in later calls
                    // [out] The answer from Server to the requested Update Rate
                    // [in] requested interface type of the group object
                    // [out] pointer to the requested interface

                    pIBHOPCServer.AddGroup(IBH_GROUP_NAME,
                        0,
                        dwRequestedUpdateRate,
                        hClientGroup,
                        hTimeBias.AddrOfPinnedObject(),
                        hDeadband.AddrOfPinnedObject(),
                        iLocale_ID,
                        out pServerGroupHandle,
                        out pRevisedUpdateRate,
                        ref iidRequiredInterface,
                        out pIBHGroup);

                    // Initialize interfaces
                    try
                    {
                        // Query interface for async group object
                        pIOPCAsyncIO2 = (IOPCAsyncIO2)pIBHGroup;
                        pIOPCGroupStateMgt = (IOPCGroupStateMgt)pIBHGroup;
                        // Query interface for group callbacks
                        pIConnectionPointContainer = (IConnectionPointContainer)pIBHGroup;
                        // Establish Callback for async operations
                        Guid iid = typeof(IOPCDataCallback).GUID;
                        pIConnectionPointContainer.FindConnectionPoint(ref iid, out pIConnectionPoint);
                        // Create a connection between the OPC servers's connection point and the clients sink.
                        pIConnectionPoint.Advise(this, out dwCookie);
                    }
                    catch (System.Exception error) // Error in "Add Group"
                    {
                        MessageBox.Show(String.Format("Error advising callbacks:-{0}", error.Message), "Add group", MessageBoxButtons.OK, MessageBoxIcon.Error);
                        return;
                    }

                    // Define items
                    ItemDefinitionArray = new OPCITEMDEF[2];

                    ItemDefinitionArray[0].szAccessPath = "";                     // No accesspath
                    ItemDefinitionArray[0].szItemID = IBH_ITEM1_NAME;             // Item #1,
                    ItemDefinitionArray[0].bActive = 1;                           // AddItem Active, OnDataChange enabled in an active group 
                    ItemDefinitionArray[0].hClient = 1;                           // Client handle , needed in callback OnDataChange
                    ItemDefinitionArray[0].dwBlobSize = 0;
                    ItemDefinitionArray[0].pBlob = IntPtr.Zero;
                    ItemDefinitionArray[0].vtRequestedDataType = (short)VarEnum.VT_UI2;  // Datatype WORD

                    ItemDefinitionArray[1].szAccessPath = "";                     // No accesspath
                    ItemDefinitionArray[1].szItemID = IBH_ITEM2_NAME;             // Item #2,
                    ItemDefinitionArray[1].bActive = 1;                           // AddItem Active, OnDataChange enabled in an active group 
                    ItemDefinitionArray[1].hClient = 2;                           // Client handle , needed in callback OnDataChange
                    ItemDefinitionArray[1].dwBlobSize = 0;
                    ItemDefinitionArray[1].pBlob = IntPtr.Zero;
                    ItemDefinitionArray[1].vtRequestedDataType = (short)VarEnum.VT_UI2;  // Datatype WORD

                    IntPtr pResults = IntPtr.Zero;
                    IntPtr pErrors = IntPtr.Zero;

                    try
                    {
                        // Add two Items
                        ((IOPCItemMgt)pIBHGroup).AddItems(2, ItemDefinitionArray, out pResults, out pErrors);

                        // Evaluate return ErrorCodes to exclude possible Errors
                         int[] errors = new int[2];
                        IntPtr position = pResults;

                        ServerItem_HandleArray = new int[2];
                        Marshal.Copy(pErrors, errors, 0, 2);

                        if (errors[0] == 0)
                        {
                            OPCITEMRESULT result = (OPCITEMRESULT)Marshal.PtrToStructure(position, typeof(OPCITEMRESULT));
                            ServerItem_HandleArray[0] = result.hServer;
                        }
                        else
                        {
                            String strErr;
                            pIBHOPCServer.GetErrorString(errors[0], iLocale_ID, out strErr);
                            MessageBox.Show(strErr, "Result for Add Item MW0", MessageBoxButtons.OK, MessageBoxIcon.Error);

                        }
                        if (errors[1] == 0)
                        {
                            position = new IntPtr(position.ToInt32() + Marshal.SizeOf(typeof(OPCITEMRESULT)));
                            OPCITEMRESULT itemres = (OPCITEMRESULT)Marshal.PtrToStructure(position, typeof(OPCITEMRESULT));
                            ServerItem_HandleArray[1] = itemres.hServer;
                        }
                        else
                        {
                            String strError;
                            pIBHOPCServer.GetErrorString(errors[1], iLocale_ID, out strError);
                            MessageBox.Show(strError, "Result for Add Item MW2", MessageBoxButtons.OK, MessageBoxIcon.Error);
                        }

                        // In case of no errors disable and enable buttons.
                        if ((errors[0] == 0) && (errors[1] == 0))
                        {
                            ReadFromOpc.Enabled = true;
                            WriteOpc.Enabled = true;
                            ExitIBHOpc.Enabled = true;
                            StartIBHOpc.Enabled = false;
                            CheckBox_Active.Enabled = true;
                            TextBox_WriteValue.Text = "0";
                        }
                    }
                    catch (System.Exception error) // Error in "Add items"
                    {
                        MessageBox.Show(error.Message, "Result for Add Items", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    }
                    finally
                    {
                        // Free memory
                        if (pResults != IntPtr.Zero)
                        {
                            Marshal.FreeCoTaskMem(pResults);
                            pResults = IntPtr.Zero;
                        }
                        if (pErrors != IntPtr.Zero)
                        {
                            Marshal.FreeCoTaskMem(pErrors);
                            pErrors = IntPtr.Zero;
                        }
                    }
                }
                catch (System.Exception error) // Error in "Add Group"
                {
                    MessageBox.Show(String.Format("Error creating OPC group:-{0}", error.Message), "Result for Add group", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
                finally
                {
                    if (hDeadband.IsAllocated) hDeadband.Free();
                    if (hTimeBias.IsAllocated) hTimeBias.Free();
                }
            }
            catch (System.Exception error) // catch for server instance creation
            {
                MessageBox.Show(String.Format("Error creating OPC Server:-{0}", error.Message), "Result for Create Server", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private void ReadFromOpc_Click(object sender, EventArgs e)
        {
            int iCancelID;
            IntPtr pErrors = IntPtr.Zero;

            if (pIOPCAsyncIO2 != null)
            {
                try
                {   // Asynchroneous read operation
                    pIOPCAsyncIO2.Read(1, ServerItem_HandleArray, iTransaction_ID + 1, out iCancelID, out pErrors);
                    int[] Errors = new int[1];
                    Marshal.Copy(pErrors, Errors, 0, 1);
                    if (Errors[0] != 0)
                    {
                        String pstrError;
                        pIBHOPCServer.GetErrorString(Errors[0], iLocale_ID, out pstrError);
                        MessageBox.Show(pstrError, "Result in Asynchroneous Read", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    }
                }
                catch (System.Exception error)
                {
                    MessageBox.Show(error.Message, "Error in Asynchroneous Read", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
            }
        }

        private void WriteOpc_Click(object sender, EventArgs e)
        {
            int iCancelID;
            IntPtr pErrors = IntPtr.Zero;
            object[] Vals = new object[1];
            Vals[0] = TextBox_WriteValue.Text;

            if (pIOPCAsyncIO2 != null)
            {
                try
                {   // Asynchroneous write operation 
                    pIOPCAsyncIO2.Write(1, ServerItem_HandleArray, Vals, iTransaction_ID + 1, out iCancelID, out pErrors);
                    int[] Errors = new int[1];
                    Marshal.Copy(pErrors, Errors, 0, 1);
                    if (Errors[0] != 0)
                    {
                        System.Exception excpt = new Exception("Error writing item");
                        Marshal.FreeCoTaskMem(pErrors);
                        pErrors = IntPtr.Zero;
                        throw excpt;
                    }
                }
                catch (System.Exception error)
                {
                    MessageBox.Show(error.Message, "Error in Asynchroneous Write", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
            }
        }

        private void ExitIBHOpc_Click(object sender, EventArgs e)
        {
            // Remove the items
            IntPtr pErrors = IntPtr.Zero;
            ((IOPCItemMgt)pIBHGroup).RemoveItems(2, ServerItem_HandleArray, out pErrors);
            
            // Unadvise the connection point.
            CheckBox_Active.Checked = false;
            if (dwCookie != 0)
            {
                pIConnectionPoint.Unadvise(dwCookie);
                dwCookie = 0;
            }

            // Free unmanaged code
            Marshal.ReleaseComObject(pIConnectionPoint);
            pIConnectionPoint = null;

            Marshal.ReleaseComObject(pIConnectionPointContainer);
            pIConnectionPointContainer = null;

            if (pIOPCAsyncIO2 != null)
            {
                Marshal.ReleaseComObject(pIOPCAsyncIO2);
                pIOPCAsyncIO2 = null;
            }
            if (pIOPCGroupStateMgt != null)
            {
                Marshal.ReleaseComObject(pIOPCGroupStateMgt);
                pIOPCGroupStateMgt = null;
            }
            if ((pIBHGroup != null) && (pIBHOPCServer != null))
            {
                pIBHOPCServer.RemoveGroup(pServerGroupHandle, 0);
            }
            if (pIBHGroup != null)
            {
                Marshal.ReleaseComObject(pIBHGroup);
                pIBHGroup = null;
            }
            if (pIBHOPCServer != null)
            {
                Marshal.ReleaseComObject(pIBHOPCServer);
                pIBHOPCServer = null;
            }

            // Reset the dialog
            TextBox_DataQuality0.Text = "";
            TextBox_DataQuality1.Text = "";
            TextBox_TimeStamp0.Text = "";
            TextBox_TimeStamp1.Text = "";
            TextBox_DataValue0.Text = "";
            TextBox_DataValue1.Text = "";
            TextBox_ReadQuality.Text = "";
            TextBox_TimeStamp.Text = "";
            TextBox_ReadVal.Text = "";
            TextBox_WriteValue.Text = "0";
            TextBox_WriteValue.Text = "";
            ReadFromOpc.Enabled = false;
            WriteOpc.Enabled = false;
            ExitIBHOpc.Enabled = false;
            StartIBHOpc.Enabled = true;
        }

        private void CheckBox_Active_CheckedChanged(object sender, EventArgs e)
        {
            IntPtr pRequestedUpdateRate = IntPtr.Zero;
            int iRevUpdateRate = 0;
            IntPtr hClientGroup = IntPtr.Zero;
            IntPtr pTimeBias = IntPtr.Zero;
            IntPtr pDeadband = IntPtr.Zero;
            IntPtr pLocale_ID = IntPtr.Zero;
            int iActive = 0;

            GCHandle hActive = GCHandle.Alloc(iActive, GCHandleType.Pinned);
            if (CheckBox_Active.Checked != true)
                hActive.Target = 0;     // Deactivate group
            else
                hActive.Target = 1;     // Activate group
            try
            {
                pIOPCGroupStateMgt.SetState(pRequestedUpdateRate, out iRevUpdateRate, hActive.AddrOfPinnedObject(), pTimeBias, pDeadband, pLocale_ID, hClientGroup);
            }
            catch (System.Exception error)
            {
                MessageBox.Show(error.Message, "Error changing the group state", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            finally
            {
                hActive.Free();
            }
        }

        // Callback IOPCDataCallback OnDataChange event handler implementation.
        public virtual void OnDataChange(   Int32 dwTransid,
                                            Int32 hGroup,
                                            Int32 hrMasterquality,
                                            Int32 hrMastererror,
                                            Int32 dwCount,
                                            int[] phClientItems,
                                            object[] pvValues,
                                            short[] pwQualities,
                                            OpcRcw.Da.FILETIME[] pftTimeStamps,
                                            int[] pErrors )
        {
            try
            {
                for (int iCount = 0; iCount < dwCount; iCount++)
                {
                    // Get Item #1 Value
                    if (pErrors[iCount] == 0)
                    {
                        if (phClientItems[iCount] == 1)// if client handel is 1 i.e. for item 0
                        {
                            // .Net Framework 2.0
                            ThreadExceptionDialog.CheckForIllegalCrossThreadCalls = false;
                            TextBox_DataValue0.Text = Convert.ToString(pvValues[iCount]);
                            TextBox_DataQuality0.Text = GetOPCQuality(pwQualities[iCount]);
                            DateTime dt = OPCTimeToDateTime(pftTimeStamps[iCount]);
                            TextBox_TimeStamp0.Text = dt.ToString();
                            // .Net Framework 2.0
                            ThreadExceptionDialog.CheckForIllegalCrossThreadCalls = true;
                        }
                        // Get Item #2 Value
                        if (phClientItems[iCount] == 2) // if client handel is 2 i.e. for item 0
                        {
                            // .Net Framework 2.0
                            ThreadExceptionDialog.CheckForIllegalCrossThreadCalls = false;
                            TextBox_DataValue1.Text = Convert.ToString(pvValues[iCount]);
                            TextBox_DataQuality1.Text = GetOPCQuality(pwQualities[iCount]);
                            DateTime dt = OPCTimeToDateTime(pftTimeStamps[iCount]);
                            TextBox_TimeStamp1.Text = dt.ToString();
                            // .Net Framework 2.0
                            ThreadExceptionDialog.CheckForIllegalCrossThreadCalls = true;
                        }
                    }
                    else
                    {
                        String strItemErr;
                        pIBHOPCServer.GetErrorString(pErrors[0], iLocale_ID, out strItemErr);
                        MessageBox.Show(strItemErr, "Error in OnDataChange", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    }
                }
            }
            catch (System.Exception exp)
            {
                MessageBox.Show(exp.Message, "Error in OnDataChange", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        // Callback IOPCDataCallback OnReadComplete event handler implementation.
        public virtual void OnReadComplete( System.Int32 dwTransid,
                                            System.Int32 hGroup,
                                            System.Int32 hrMasterquality,
                                            System.Int32 hrMastererror,
                                            System.Int32 dwCount,
                                            int[] phClientItems,
                                            object[] pvValues,
                                            short[] pwQualities,
                                            OpcRcw.Da.FILETIME[] pftTimeStamps,
                                            int[] pErrors )
        {
            try
            {
                if (pErrors[0] == 0)
                {
                    // .Net Framework 2.0
                    ThreadExceptionDialog.CheckForIllegalCrossThreadCalls = false;
                    TextBox_ReadVal.Text = String.Format("{0}", pvValues[0]);
                    TextBox_ReadQuality.Text = GetOPCQuality(pwQualities[0]);
                    DateTime dt = OPCTimeToDateTime(pftTimeStamps[0]);
                    TextBox_TimeStamp.Text = dt.ToString();
                    // .Net Framework 2.0
                    ThreadExceptionDialog.CheckForIllegalCrossThreadCalls = true;
                }
                else
                {
                    String strRes = "";
                    pIBHOPCServer.GetErrorString(pErrors[0], iLocale_ID, out strRes);
                    MessageBox.Show(strRes, "Error in OnReadComplete", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
            }
            catch (System.Exception exp)
            {
                MessageBox.Show(exp.Message, "Error in OnReadComplete", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        // Callback IOPCDataCallback OnWriteComplete event handler implementation.
        public virtual void OnWriteComplete(    System.Int32 dwTransid,
                                                System.Int32 hGroup,
                                                System.Int32 hrMastererr,
                                                System.Int32 dwCount,
                                                int[] pClienthandles,
                                                int[] pErrors )
        {
            // .Net Framework 2.0
            ThreadExceptionDialog.CheckForIllegalCrossThreadCalls = false;
            String strRes = "";
            pIBHOPCServer.GetErrorString(pErrors[0], iLocale_ID, out strRes);
            TextBox_WriteResult.Text = strRes;
            // .Net Framework 2.0
            ThreadExceptionDialog.CheckForIllegalCrossThreadCalls = true;
        }

        // Callback IOPCDataCallback OnCancelComplete event handler implementation.
        public virtual void OnCancelComplete(System.Int32 dwTransid, System.Int32 hGroup)
        {
            // Not needed
        }

        private void IBHOpc_CSharp_Load(object sender, System.EventArgs e)
        {
            ReadFromOpc.Enabled = false;
            WriteOpc.Enabled = false;
            ExitIBHOpc.Enabled = false;
            TextBox_WriteValue.Text = "0";
        }

        /// Convert OPC Time FILETIME to DateTime
        private DateTime OPCTimeToDateTime(OpcRcw.Da.FILETIME ft)
        {
            long highbuf = (long)ft.dwHighDateTime;
            long buf;

            if (ft.dwLowDateTime >= 0)
                buf = (highbuf << 32) + ft.dwLowDateTime;
            else
                buf = ((highbuf + 1) << 32) + ft.dwLowDateTime;

            return DateTime.FromFileTimeUtc(buf);
        }

        // OPC foundation quality values.
        private String GetOPCQuality(long wQuality)
        {
            String strOPCQuality = "";
            switch (wQuality)
            {
                case Qualities.OPC_QUALITY_GOOD:
                    strOPCQuality = "Good";
                    break;
                case Qualities.OPC_QUALITY_BAD:
                    strOPCQuality = "Bad";
                    break;
                case Qualities.OPC_QUALITY_CONFIG_ERROR:
                    strOPCQuality = "BadConfigurationError";
                    break;
                case Qualities.OPC_QUALITY_NOT_CONNECTED:
                    strOPCQuality = "BadNotConnected";
                    break;
                case Qualities.OPC_QUALITY_DEVICE_FAILURE:
                    strOPCQuality = "BadDeviceFailure";
                    break;
                case Qualities.OPC_QUALITY_SENSOR_FAILURE:
                    strOPCQuality = "BadSensorFailure";
                    break;
                case Qualities.OPC_QUALITY_COMM_FAILURE:
                    strOPCQuality = "BadCommFailure";
                    break;
                case Qualities.OPC_QUALITY_OUT_OF_SERVICE:
                    strOPCQuality = "BadOutOfService";
                    break;
                case Qualities.OPC_QUALITY_WAITING_FOR_INITIAL_DATA:
                    strOPCQuality = "BadWaitingForInitialData";
                    break;
                case Qualities.OPC_QUALITY_EGU_EXCEEDED:
                    strOPCQuality = "UncertainEGUExceeded";
                    break;
                case Qualities.OPC_QUALITY_SUB_NORMAL:
                    strOPCQuality = "UncertainSubNormal";
                    break;
                default:
                    strOPCQuality = "Not handled";
                    break;
            }
            return strOPCQuality;
        }
    }
}