Option Strict Off
Option Explicit On

Imports System
Imports System.Reflection
Imports System.Runtime.CompilerServices
Imports System.Runtime.InteropServices

' To get the imports, add dependencys to the files 
' OpcRcwComn.dll and OpcRcw.Da.dll typically located in the
' <Program Files>\<Common Files>\OPC Foundation\Bin folder
Imports OpcRcw.Comn
Imports OpcRcw.Da

Public Class IBHOPCSample

    Implements IOPCDataCallback

    Dim m_IBHOPCServer As IOPCServer = Nothing
    Dim m_IBHOPCGroup As IOPCGroupStateMgt = Nothing
    Dim m_IBHOPCItem As IOPCItemMgt = Nothing
    Dim m_AsyncIO2 As IOPCAsyncIO2 = Nothing

    Dim m_ConnectionContainer As IConnectionPointContainer = Nothing
    Dim m_ConnectionPoint As IConnectionPoint = Nothing
    Dim m_Callback As IOPCDataCallback = Nothing
    Dim m_dwCookie As Int32 = 0

    'Language code &H407 : English
    Const iLocale_ID As Integer = &H407

    Dim iTransaction As Integer = 0
    Dim iServerGroup As Integer = 0
    Dim iClientGrouphandle As Integer = 0
    Dim iServerHandle1 As Integer = 0
    Dim iServerHandle2 As Integer = 0
    Dim iActiveGroup As Boolean = False


    Private Sub StartIBHOpc_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles StartIBHOpc.Click

        Try
            ' Use the IBH OPC Server ProgID
            Dim typeofOPCserver As Type = Type.GetTypeFromProgID("IBHsoftec.IBHOPC.DA.1")

            ' Create the IBH OPC Server object
            m_IBHOPCServer = Activator.CreateInstance(typeofOPCserver)

            Dim RevisedUpdateRate As Integer
            Dim TimeBias As Int32 = 0
            Dim Deadband As Double = 0
            Dim hTimeBias As GCHandle = GCHandle.Alloc(TimeBias, GCHandleType.Pinned)
            Dim hDeadband As GCHandle = GCHandle.Alloc(TimeBias, GCHandleType.Pinned)

            iClientGrouphandle = 10
            ' Add the group "MyIBHOPCGroup"
            ' Parameter as following:
            ' [in] not active, so no OnDataChange callback
            ' [in] Request this Update Rate from Server
            ' [in] Client Handle, not necessary in this sample
            ' [in] No time interval to system UTC time
            ' [in] No Deadband, so all data changes are reported
            ' [in] Server uses english language to for text values
            ' [out] Server handle to identify this group in later calls
            ' [out] The answer from Server to the requested Update Rate
            ' [in] requested interface type of the group object
            ' [out] pointer to the requested interface
            m_IBHOPCServer.AddGroup( _
                    "MyIBHOPCGroup", _
                    0, _
                    1000, _
                    iClientGrouphandle, _
                    hTimeBias.AddrOfPinnedObject(), _
                    hDeadband.AddrOfPinnedObject(), _
                    iLocale_ID, _
                    iServerGroup, _
                    RevisedUpdateRate, _
                    GetType(IOPCGroupStateMgt).GUID, _
                    m_IBHOPCGroup)
            m_IBHOPCItem = m_IBHOPCGroup

            ' Define items
            Dim itemdefs(2) As OPCITEMDEF
            itemdefs(0) = New OPCITEMDEF
            ' No accesspath
            itemdefs(0).szAccessPath = ""
            ' ItemID
            itemdefs(0).szItemID = "Steuerung1.MW0"
            ' AddItem Active, OnDataChange enabled in an active group 
            itemdefs(0).bActive = 1
            ' Client handle , needed in this sample in callback OnDataChange
            itemdefs(0).hClient = 1
            itemdefs(0).vtRequestedDataType = VarEnum.VT_UI2
            itemdefs(0).dwBlobSize = 0
            itemdefs(0).pBlob = IntPtr.Zero

            itemdefs(1) = New OPCITEMDEF
            itemdefs(1).hClient = 2
            itemdefs(1).szItemID = "Steuerung1.MW2"
            itemdefs(1).szAccessPath = ""
            itemdefs(1).bActive = 1
            itemdefs(1).vtRequestedDataType = VarEnum.VT_UI2
            itemdefs(1).dwBlobSize = 0
            itemdefs(1).pBlob = IntPtr.Zero

            Dim Results As IntPtr = IntPtr.Zero
            Dim pErrors As IntPtr = IntPtr.Zero

            ' Add two Items
            m_IBHOPCItem.AddItems(2, _
                                    itemdefs, _
                                    Results, _
                                    pErrors)


            'Evaluate return ErrorCodes to exclude possible Errors
            Dim errors(2) As Integer
            Marshal.Copy(pErrors, errors, 0, 2)
            Dim result As OPCITEMRESULT
            Dim pos As IntPtr = Results
            If errors(0) = 0 Then
                ' First Item was added succesfully
                result = Marshal.PtrToStructure(pos, GetType(OPCITEMRESULT))
                iServerHandle1 = result.hServer
            Else
                ' First Item was not added
                ' Raise Exception
                Dim ex As Exception = New Exception("Error adding Item")
                Throw ex
            End If

            If errors(1) = 0 Then
                ' Second Item was added succesfully
                pos = New IntPtr(pos.ToInt32() + Marshal.SizeOf(GetType(OPCITEMRESULT)))
                result = Marshal.PtrToStructure(pos, GetType(OPCITEMRESULT))
                iServerHandle2 = result.hServer
            Else
                ' Second Item was not added
                ' Raise Exception
                Dim ex As Exception = New Exception("Error adding Item")
                Throw ex
            End If

            ' Free allocated COM-ressouces
            Marshal.FreeCoTaskMem(Results)
            Marshal.FreeCoTaskMem(pErrors)



            ' Query interface for async calls on group object
            m_AsyncIO2 = m_IBHOPCGroup

            ' Activate Group ?
            iActiveGroup = False

            ' Query interface for callbacks on group object
            m_ConnectionContainer = m_IBHOPCGroup

            ' Establish Callback for all async operations
            m_ConnectionContainer.FindConnectionPoint(GetType(IOPCDataCallback).GUID, m_ConnectionPoint)

            ' Creates a connection between the OPC servers's connection point and 
            ' this client's sink (the callback object). 
            m_ConnectionPoint.Advise(Me, m_dwCookie)

            ' Success, enable communication
            StartIBHOpc.Enabled = False
            ReadFromOpc.Enabled = True
            WriteOpc.Enabled = True
            ExitIBHOpc.Enabled = True
            CheckBox_Active.Enabled = True


        Catch ex As Exception
            MessageBox.Show(ex.ToString(), "Exception", MessageBoxButtons.OK, MessageBoxIcon.Error)
        End Try

    End Sub

    Private Sub ReadFromOpc_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles ReadFromOpc.Click

        Try
            Dim CancelID As Integer
            Dim pServer(1) As Integer
            ' Select item by server handle, received at AddItem
            pServer(0) = iServerHandle1
            Dim pItemValues As IntPtr = IntPtr.Zero, pErrors As IntPtr = IntPtr.Zero
            iTransaction += 1
            m_AsyncIO2.Read(1, _
                pServer, _
                iTransaction, _
                CancelID, _
                pErrors)

            ' Check for errors
            Dim errors(1) As Integer
            Marshal.Copy(pErrors, errors, 0, 1)
            If errors(0) <> 0 Then
                ' Errors occured - raise Exception
                Dim ex As Exception = New Exception("Read Error")
                Throw ex
            End If


        Catch ex As Exception
            MessageBox.Show(ex.ToString(), "Exception", MessageBoxButtons.OK, MessageBoxIcon.Error)
        End Try

        Exit Sub

    End Sub

    Private Sub WriteOpc_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles WriteOpc.Click

        Try

            Dim phServer(1) As Integer
            phServer(0) = iServerHandle1
            Dim phItemValues(1) As Object
            phItemValues(0) = TextBox_WriteValue.Text
            Dim dwCancelID As Integer
            Dim pErrors As IntPtr = IntPtr.Zero

            iTransaction += 1
            ' Select item by server handle, received at AddItem
            m_AsyncIO2.Write(1, _
                 phServer, _
                 phItemValues, _
                 iTransaction, _
                 dwCancelID, _
                 pErrors)

            'check if an error occured
            Dim errors(1) As Integer
            Marshal.Copy(pErrors, errors, 0, 1)
            If errors(0) <> 0 Then
                ' Errors occured - raise Exception
                TextBox_WriteResult.Text = "OPC_Quality_Bad"
                Dim ex As Exception = New Exception("Write Error")
                Throw ex
            End If

        Catch ex As Exception
            MessageBox.Show(ex.ToString(), "Exception", MessageBoxButtons.OK, MessageBoxIcon.Error)
        End Try

    End Sub

    Private Sub ExitIBHOpc_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles ExitIBHOpc.Click

        Try
            CheckBox_Active.Checked = False

            ' Removes items
            Dim pItems(2) As Integer
            ' Select item by server handle, received at AddItem
            pItems(0) = iServerHandle1
            pItems(1) = iServerHandle2

            Dim pErrors As IntPtr = IntPtr.Zero

            m_IBHOPCItem.RemoveItems(2, _
                                pItems, _
                                pErrors)

            'Check for Errors
            Dim errors(2) As Integer
            Marshal.Copy(pErrors, errors, 0, 2)
            If errors(0) = 0 Then
                ' Item deleted succesfully
                iServerHandle1 = 0
            Else
                ' Item was not added, raise Exception
                Dim ex As Exception = New Exception("Error while deleting first Item")
                Throw ex
            End If

            Marshal.FreeCoTaskMem(pErrors)

            ' Unadvise and remove the callback
            m_ConnectionPoint.Unadvise(m_dwCookie)
            Marshal.ReleaseComObject(m_ConnectionPoint)
            m_ConnectionPoint = Nothing
            m_ConnectionContainer = Nothing

            ' Remove Group
            m_IBHOPCServer.RemoveGroup(iServerGroup, True)

            ' Release all interfaces
            m_AsyncIO2 = Nothing
            m_IBHOPCItem = Nothing
            m_IBHOPCGroup = Nothing
            Marshal.ReleaseComObject(m_IBHOPCServer)
            m_IBHOPCServer = Nothing
            ' The Server is terminated now

            ' Cleanup
            TextBox_ReadVal.Text = ""
            TextBox_ReadQuality.Text = ""
            TextBox_TimeStamp.Text = ""
            TextBox_DataValue0.Text = ""
            TextBox_DataQuality0.Text = ""
            TextBox_TimeStamp0.Text = ""
            TextBox_DataValue1.Text = ""
            TextBox_DataQuality1.Text = ""
            TextBox_TimeStamp1.Text = ""
            TextBox_WriteResult.Text = ""
            TextBox_WriteValue.Text = "0"

            StartIBHOpc.Enabled = True
            ReadFromOpc.Enabled = False
            WriteOpc.Enabled = False
            ExitIBHOpc.Enabled = False
            CheckBox_Active.Enabled = False
            TextBox_WriteValue.ReadOnly = True
        Catch ex As Exception
            MessageBox.Show(ex.ToString(), "Exception", MessageBoxButtons.OK, MessageBoxIcon.Error)
        End Try

        Exit Sub

    End Sub

    Private Sub CheckBox_Active_CheckedChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles CheckBox_Active.CheckedChanged

        Dim RequestedUpdateRate As Integer = 1000
        Dim Active As Integer = 0
        Dim Loc As Integer = &H409

        Dim pTimeBias As IntPtr = IntPtr.Zero
        Dim pDeadband As IntPtr = IntPtr.Zero
        Dim RevisedUpdateRate As Integer

        ' Access unmanged memory
        Dim hRequestedUpdateRate As GCHandle = GCHandle.Alloc(RequestedUpdateRate, GCHandleType.Pinned)
        Dim hLoc As GCHandle = GCHandle.Alloc(Loc, GCHandleType.Pinned)
        Dim hActive As GCHandle = GCHandle.Alloc(Active, GCHandleType.Pinned)
        Dim hClientGroup As GCHandle = GCHandle.Alloc(iClientGrouphandle, GCHandleType.Pinned)

        Try
            If Not iActiveGroup Then
                ' Activate group
                iActiveGroup = True
                hActive.Target = 1
                m_IBHOPCGroup.SetState(hRequestedUpdateRate.AddrOfPinnedObject(), _
                    RevisedUpdateRate, _
                    hActive.AddrOfPinnedObject(), _
                    pTimeBias, _
                    pDeadband, _
                    hLoc.AddrOfPinnedObject(), _
                    hClientGroup.AddrOfPinnedObject())
            Else
                ' Deactivate group
                iActiveGroup = False
                hActive.Target = 0
                m_IBHOPCGroup.SetState(hRequestedUpdateRate.AddrOfPinnedObject(), _
                    RevisedUpdateRate, _
                    hActive.AddrOfPinnedObject(), _
                    pTimeBias, _
                    pDeadband, _
                    hLoc.AddrOfPinnedObject(), _
                    hClientGroup.AddrOfPinnedObject())
                ' Clear previously shown values
                TextBox_DataValue0.Text = ""
                TextBox_DataQuality0.Text = ""
                TextBox_TimeStamp0.Text = ""
                TextBox_DataValue1.Text = ""
                TextBox_DataQuality1.Text = ""
                TextBox_TimeStamp1.Text = ""
            End If

        Catch ex As Exception
            MessageBox.Show(ex.ToString(), "Exception", MessageBoxButtons.OK, MessageBoxIcon.Error)
        Finally
            ' Free unmangaged memory
            If hActive.IsAllocated Then hActive.Free()
            If hLoc.IsAllocated Then hLoc.Free()
            If hRequestedUpdateRate.IsAllocated Then hRequestedUpdateRate.Free()
            If hClientGroup.IsAllocated Then hClientGroup.Free()
        End Try

    End Sub

    Public Sub OnCancelComplete(ByVal dwTransid As Integer, ByVal hGroup As Integer) Implements OpcRcw.Da.IOPCDataCallback.OnCancelComplete

    End Sub

    Private Function GetOPCQualityText(ByVal qnr As Integer) As String
        Const OPC_QUALITY_MASK = &HC0
        Const OPC_STATUS_MASK = &HFC
        Const OPC_LIMIT_MASK = &H3
        Const OPC_QUALITY_BAD = &H0
        Const OPC_QUALITY_UNCERTAIN = &H40
        Const OPC_QUALITY_GOOD = &HC0
        Const OPC_QUALITY_CONFIG_ERROR = &H4
        Const OPC_QUALITY_NOT_CONNECTED = &H8
        Const OPC_QUALITY_DEVICE_FAILURE = &HC
        Const OPC_QUALITY_SENSOR_FAILURE = &H10
        Const OPC_QUALITY_LAST_KNOWN = &H14
        Const OPC_QUALITY_COMM_FAILURE = &H18
        Const OPC_QUALITY_OUT_OF_SERVICE = &H1C
        Const OPC_QUALITY_LAST_USABLE = &H44
        Const OPC_QUALITY_SENSOR_CAL = &H50
        Const OPC_QUALITY_EGU_EXCEEDED = &H54
        Const OPC_QUALITY_SUB_NORMAL = &H58
        Const OPC_QUALITY_LOCAL_OVERRIDE = &HD8
        Const OPC_LIMIT_OK = &H0
        Const OPC_LIMIT_LOW = &H1
        Const OPC_LIMIT_HIGH = &H2
        Const OPC_LIMIT_CONST = &H3

        Dim qstr As String = ""

        Select Case qnr

            Case OPC_QUALITY_BAD
                qstr = "BAD"
            Case OPC_QUALITY_UNCERTAIN
                qstr = "UNCERTAIN"
            Case OPC_QUALITY_GOOD
                qstr = "GOOD"
            Case OPC_QUALITY_NOT_CONNECTED
                qstr = "NOT_CONNECTED"
            Case OPC_QUALITY_DEVICE_FAILURE
                qstr = "DEVICE_FAILURE"
            Case OPC_QUALITY_SENSOR_FAILURE
                qstr = "SENSOR_FAILURE"
            Case OPC_QUALITY_LAST_KNOWN
                qstr = "LAST_KNOWN"
            Case OPC_QUALITY_COMM_FAILURE
                qstr = "COMM_FAILURE"
            Case OPC_QUALITY_OUT_OF_SERVICE
                qstr = "OUT_OF_SERVICE"
            Case OPC_QUALITY_LAST_USABLE
                qstr = "LAST_USABLE"
            Case OPC_QUALITY_SENSOR_CAL
                qstr = "SENSOR_CAL"
            Case OPC_QUALITY_EGU_EXCEEDED
                qstr = "EGU_EXCEEDED"
            Case OPC_QUALITY_SUB_NORMAL
                qstr = "SUB_NORMAL"
            Case OPC_QUALITY_LOCAL_OVERRIDE
                qstr = "LOCAL_OVERRIDE"
            Case Else
                qstr = "UNKNOWN ERROR"
        End Select
        Return qstr
    End Function


    Public Function OPCTimeToDateTime(ByVal ft As OpcRcw.Da.FILETIME) As DateTime
        ' FILETIME to DateTime
        Dim High As Long = ft.dwHighDateTime
        Dim lng As Long = (High << 32) + ft.dwLowDateTime
        Return DateTime.FromFileTimeUtc(lng)
    End Function


    Public Sub OnDataChange(ByVal dwTransid As Integer, ByVal hGroup As Integer, ByVal hrMasterquality As Integer, ByVal hrMastererror As Integer, ByVal dwCount As Integer, ByVal phClientItems() As Integer, ByVal pvValues() As Object, ByVal pwQualities() As Short, ByVal pftTimeStamps() As OpcRcw.Da.FILETIME, ByVal pErrors() As Integer) Implements OpcRcw.Da.IOPCDataCallback.OnDataChange

        Try
            If dwCount >= 1 Then
                ' Data received
                If pErrors(0) = 0 Then
                    ' First item without errors
                    Dim dt As DateTime = OPCTimeToDateTime(pftTimeStamps(0))
                    ' Display first value
                    If phClientItems(0) = 1 Then
                        ' First value corresponds with upper fields
                        '.Net Framework 2.0 
                        ThreadExceptionDialog.CheckForIllegalCrossThreadCalls = False
                        TextBox_DataValue0.Text = Convert.ToString(pvValues(0))
                        TextBox_DataQuality0.Text = GetOPCQualityText(pwQualities(0))
                        TextBox_TimeStamp0.Text = Convert.ToString(dt)
                        '.Net Framework 2.0 
                        ThreadExceptionDialog.CheckForIllegalCrossThreadCalls = True
                    Else
                        ' First value corresponds with lower fields
                        '.Net Framework 2.0 
                        ThreadExceptionDialog.CheckForIllegalCrossThreadCalls = False
                        TextBox_DataValue1.Text = Convert.ToString(pvValues(0))
                        TextBox_DataQuality1.Text = GetOPCQualityText(pwQualities(0))
                        TextBox_TimeStamp1.Text = Convert.ToString(dt)
                        '.Net Framework 2.0 
                        ThreadExceptionDialog.CheckForIllegalCrossThreadCalls = True
                    End If
                Else
                    Dim ErrorString As String = "Read Error"
                    MsgBox(ErrorString, MsgBoxStyle.Critical, "Error AsyncReadComplete()")
                End If
            End If

            If dwCount >= 2 Then
                ' Values received
                ' Display second Value
                If pErrors(1) = 0 Then
                    ' Second value received without error
                    Dim dt As DateTime = OPCTimeToDateTime(pftTimeStamps(1))
                    If phClientItems(1) = 2 Then
                        ' First value corresponds with upper fields
                        '.Net Framework 2.0 
                        ThreadExceptionDialog.CheckForIllegalCrossThreadCalls = False
                        TextBox_DataValue1.Text = Convert.ToString(pvValues(1))
                        TextBox_DataQuality1.Text = GetOPCQualityText(pwQualities(1))
                        TextBox_TimeStamp1.Text = Convert.ToString(dt)
                        '.Net Framework 2.0 
                        ThreadExceptionDialog.CheckForIllegalCrossThreadCalls = True
                    Else
                        ' First value corresponds with lower fields
                        '.Net Framework 2.0 
                        ThreadExceptionDialog.CheckForIllegalCrossThreadCalls = False
                        TextBox_DataValue0.Text = Convert.ToString(pvValues(1))
                        TextBox_DataQuality0.Text = GetOPCQualityText(pwQualities(1))
                        TextBox_TimeStamp0.Text = Convert.ToString(dt)
                        '.Net Framework 2.0 
                        ThreadExceptionDialog.CheckForIllegalCrossThreadCalls = True
                    End If
                Else
                    Dim ErrorString As String = "Read Error"
                    MsgBox(ErrorString, MsgBoxStyle.Critical, "Error AsyncReadComplete()")
                End If
            End If

        Catch e As Exception
            Dim msg As String = e.Message
            MsgBox(msg, MsgBoxStyle.Critical)
        End Try

    End Sub

    Public Sub OnReadComplete(ByVal dwTransid As Integer, ByVal hGroup As Integer, ByVal hrMasterquality As Integer, ByVal hrMastererror As Integer, ByVal dwCount As Integer, ByVal phClientItems() As Integer, ByVal pvValues() As Object, ByVal pwQualities() As Short, ByVal pftTimeStamps() As OpcRcw.Da.FILETIME, ByVal pErrors() As Integer) Implements OpcRcw.Da.IOPCDataCallback.OnReadComplete

        Try
            If pErrors(0) = 0 Then
                ' Display value
                '.Net Framework 2.0
                ThreadExceptionDialog.CheckForIllegalCrossThreadCalls = False
                TextBox_ReadVal.Text = pvValues(0)
                TextBox_ReadQuality.Text = GetOPCQualityText(pwQualities(0))
                Dim dt As DateTime = OPCTimeToDateTime(pftTimeStamps(0))
                TextBox_TimeStamp.Text = Convert.ToString(dt)
                '.Net Framework 2.0 
                ThreadExceptionDialog.CheckForIllegalCrossThreadCalls = True
            Else
                Dim ErrorString As String = "Read Error"
                MsgBox(ErrorString, MsgBoxStyle.Critical, "Error AsyncReadComplete()")
            End If

        Catch e As Exception
            Dim msg As String = e.Message
            MsgBox(msg, MsgBoxStyle.Critical)
        End Try

    End Sub

    Public Sub OnWriteComplete(ByVal dwTransid As Integer, ByVal hGroup As Integer, ByVal hrMastererr As Integer, ByVal dwCount As Integer, ByVal pClienthandles() As Integer, ByVal pErrors() As Integer) Implements OpcRcw.Da.IOPCDataCallback.OnWriteComplete

        Try
            Dim ErrorString As String = ""

            If dwCount = 1 Then
                '.Net Framework 2.0 
                ThreadExceptionDialog.CheckForIllegalCrossThreadCalls = False
                m_IBHOPCServer.GetErrorString(pErrors(0), iLocale_ID, ErrorString)
                TextBox_WriteResult.Text = ErrorString
                '.Net Framework 2.0 
                ThreadExceptionDialog.CheckForIllegalCrossThreadCalls = True
            End If
        Catch e As Exception
            Dim msg As String = e.Message
            MsgBox(msg, MsgBoxStyle.Critical)
        End Try

    End Sub

    Private Sub TextBox_WriteValue_TextChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles TextBox_WriteValue.TextChanged

        WriteOpc.Enabled = True
        TextBox_WriteResult.Text = ""

    End Sub

    Private Sub IBHOPCSample_Load(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.Load

        TextBox_WriteValue.Text = "0"

    End Sub
End Class
